/*
 * SPI Flash Core
 *
 * Copyright (C) 2015 Jagan Teki <jteki@openedev.com>
 * Copyright (C) 2013 Jagannadha Sutradharudu Teki, Xilinx Inc.
 * Copyright (C) 2010 Reinhard Meyer, EMK Elektronik
 * Copyright (C) 2008 Atmel Corporation
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <errno.h>
#include <malloc.h>
#include <mapmem.h>
#include <spi.h>
#include <spi_flash.h>
#include <linux/log2.h>
#include <dma.h>

#include "sf_internal.h"

DECLARE_GLOBAL_DATA_PTR;

static void spi_flash_addr(const struct spi_flash *flash, u32 addr, u8 *cmd)
{
	u8 shift = flash->addr_width * 8;

	/* cmd[0] is actual command */
	cmd[1] = addr >> (shift -  8);
	cmd[2] = addr >> (shift - 16);
	cmd[3] = addr >> (shift - 24);
	cmd[4] = addr >> (shift - 32);
}

static int read_sr(struct spi_flash *flash, u8 *rs)
{
	int ret;

	ret = spi_flash_read_reg(flash, CMD_READ_STATUS, 1, rs);
	if (ret < 0) {
		debug("SF: fail to read status register\n");
		return ret;
	}

	return 0;
}

static int read_fsr(struct spi_flash *flash, u8 *fsr)
{
	int ret;

	ret = spi_flash_read_reg(flash, CMD_FLAG_STATUS, 1, fsr);
	if (ret < 0) {
		debug("SF: fail to read flag status register\n");
		return ret;
	}

	return 0;
}

static int write_sr(struct spi_flash *flash, u8 ws)
{
	int ret;

	ret = spi_flash_update_reg(flash, CMD_WRITE_STATUS, 1, &ws);
	if (ret < 0) {
		debug("SF: fail to write status register\n");
		return ret;
	}

	return 0;
}

#if defined(CONFIG_SPI_FLASH_SPANSION) || defined(CONFIG_SPI_FLASH_WINBOND)
static int read_cr(struct spi_flash *flash, u8 *rc)
{
	int ret;

	ret = spi_flash_read_reg(flash, CMD_READ_CONFIG, 1, rc);
	if (ret < 0) {
		debug("SF: fail to read config register\n");
		return ret;
	}

	return 0;
}

static int write_cr(struct spi_flash *flash, u8 wc)
{
	u8 data[2];
	int ret;

	ret = read_sr(flash, &data[0]);
	if (ret < 0)
		return ret;

	data[1] = wc;
	ret = spi_flash_update_reg(flash, CMD_WRITE_STATUS, 2, data);
	if (ret) {
		debug("SF: fail to write config register\n");
		return ret;
	}

	return 0;
}
#endif

#ifdef CONFIG_SPI_FLASH_BAR
static int spi_flash_write_bar(struct spi_flash *flash, u32 offset)
{
	u8 bank_sel;
	int ret;

	bank_sel = offset / (SPI_FLASH_16MB_BOUN << flash->shift);
	if (bank_sel == flash->bank_curr)
		goto bar_end;

	ret = spi_flash_update_reg(flash, flash->bank_write_cmd, 1, &bank_sel);
	if (ret < 0) {
		debug("SF: fail to write bank register\n");
		return ret;
	}

bar_end:
	flash->bank_curr = bank_sel;
	return flash->bank_curr;
}

static int spi_flash_read_bar(struct spi_flash *flash, u8 idcode0)
{
	u8 curr_bank = 0;
	int ret;

	if (flash->size <= SPI_FLASH_16MB_BOUN)
		goto bar_end;

	switch (idcode0) {
	case SPI_FLASH_CFI_MFR_SPANSION:
		flash->bank_read_cmd = CMD_BANKADDR_BRRD;
		flash->bank_write_cmd = CMD_BANKADDR_BRWR;
		break;
	default:
		flash->bank_read_cmd = CMD_EXTNADDR_RDEAR;
		flash->bank_write_cmd = CMD_EXTNADDR_WREAR;
	}

	ret = spi_flash_read_reg(flash, flash->bank_read_cmd, &curr_bank, 1);
	if (ret) {
		debug("SF: fail to read bank addr register\n");
		return ret;
	}

bar_end:
	flash->bank_curr = curr_bank;
	return 0;
}
#endif

#ifdef CONFIG_SPI_FLASH_4B_OPCODES
struct address_entry {
	u8	src_opcode;
	u8	dst_opcode;
};

static u8 spi_flash_convert_opcode(u8 opcode,
				   const struct address_entry *entries,
				   size_t num_entries)
{
	int b, e;

	b = 0;
	e = num_entries - 1;
	while (b <= e) {
		int m = (b + e) >> 1;
		const struct address_entry *entry = &entries[m];

		if (opcode == entry->src_opcode)
			return entry->dst_opcode;

		if (opcode < entry->src_opcode)
			e = m - 1;
		else
			b = m + 1;
	}

	/* No convertion found */
	return opcode;
}

static u8 spi_flash_3to4_opcode(u8 opcode)
{
	/* MUST be sorted by 3byte opcode */
#define ENTRY_3TO4(_opcode)	{ _opcode, _opcode##_4B }
	static const struct address_entry spi_flash_3to4_table[] = {
		ENTRY_3TO4(CMD_PAGE_PROGRAM),		/* 0x02 */
		ENTRY_3TO4(CMD_READ_ARRAY_SLOW),	/* 0x03 */
		ENTRY_3TO4(CMD_READ_ARRAY_FAST),	/* 0x0b */
		ENTRY_3TO4(CMD_ERASE_4K),		/* 0x20 */
		ENTRY_3TO4(CMD_QUAD_PAGE_PROGRAM),	/* 0x32 */
		ENTRY_3TO4(CMD_QUAD_PAGE_PROGRAM_MXIC),	/* 0x38 */
		ENTRY_3TO4(CMD_READ_DUAL_OUTPUT_FAST),	/* 0x3b */
		ENTRY_3TO4(CMD_ERASE_32K),		/* 0x52 */
		ENTRY_3TO4(CMD_READ_QUAD_OUTPUT_FAST),	/* 0x6b */
		ENTRY_3TO4(CMD_READ_DUAL_IO_FAST),	/* 0xbb */
		ENTRY_3TO4(CMD_ERASE_64K),		/* 0xd8 */
		ENTRY_3TO4(CMD_READ_QUAD_IO_FAST),	/* 0xeb */
	};
#undef ENTRY_3TO4

	return spi_flash_convert_opcode(opcode,
					spi_flash_3to4_table,
					ARRAY_SIZE(spi_flash_3to4_table));
}
#endif /* CONFIG_SPI_FLASH_4B_OPCODES */

#ifdef CONFIG_SF_DUAL_FLASH
static void spi_flash_dual(struct spi_flash *flash, u32 *addr)
{
	struct spi_slave *spi = flash->spi;

	switch (flash->dual_flash) {
	case SF_DUAL_STACKED_FLASH:
		if (*addr >= (flash->size >> 1)) {
			*addr -= flash->size >> 1;
			spi->flags |= SPI_XFER_U_PAGE;
		} else {
			spi->flags &= ~SPI_XFER_U_PAGE;
		}
		break;
	case SF_DUAL_PARALLEL_FLASH:
		*addr >>= flash->shift;
		break;
	default:
		debug("SF: Unsupported dual_flash=%d\n", flash->dual_flash);
		break;
	}
}
#endif

static int spi_flash_sr_ready(struct spi_flash *flash)
{
	u8 sr;
	int ret;

	ret = read_sr(flash, &sr);
	if (ret < 0)
		return ret;

	return !(sr & STATUS_WIP);
}

static int spi_flash_fsr_ready(struct spi_flash *flash)
{
	u8 fsr;
	int ret;

	ret = read_fsr(flash, &fsr);
	if (ret < 0)
		return ret;

	return fsr & STATUS_PEC;
}

static int spi_flash_ready(struct spi_flash *flash)
{
	int sr, fsr;

	sr = spi_flash_sr_ready(flash);
	if (sr < 0)
		return sr;

	fsr = 1;
	if (flash->flags & SNOR_F_USE_FSR) {
		fsr = spi_flash_fsr_ready(flash);
		if (fsr < 0)
			return fsr;
	}

	return sr && fsr;
}

static int spi_flash_cmd_wait_ready(struct spi_flash *flash,
					unsigned long timeout)
{
	int timebase, ret;

	timebase = get_timer(0);

	while (get_timer(timebase) < timeout) {
		ret = spi_flash_ready(flash);
		if (ret < 0)
			return ret;
		if (ret)
			return 0;
	}

	printf("SF: Timeout!\n");

	return -ETIMEDOUT;
}

int spi_flash_wait_ready(struct spi_flash *flash)
{
	return spi_flash_cmd_wait_ready(flash, SPI_FLASH_PROG_TIMEOUT);
}

int spi_flash_update_reg(struct spi_flash *flash, u8 opcode,
			 size_t len, const void *buf)
{
	int ret;

	ret = spi_flash_cmd_write_enable(flash);
	if (ret < 0) {
		debug("SF: enabling write failed\n");
		return ret;
	}

	ret = spi_flash_write_reg(flash, opcode, len, buf);
	if (ret < 0) {
		debug("SF: write cmd failed\n");
		return ret;
	}

	ret = spi_flash_wait_ready(flash);
	if (ret < 0) {
		debug("SF: write register timed out\n");
		return ret;
	}

	return 0;
}

int spi_flash_cmd_write_reg_ops(struct spi_flash *flash, u8 opcode, size_t len,
				const void *buf)
{
	struct spi_slave *spi = flash->spi;

	return spi_flash_cmd_write(spi, &opcode, 1, buf, len);
}

int spi_flash_erase_alg(struct spi_flash *flash, u32 offset, size_t len,
			spi_flash_erase_fn erase_fn)
{
	struct spi_slave *spi = flash->spi;
	u32 erase_size, erase_addr;
	int ret = -1;

	erase_size = flash->erase_size;
	if (offset % erase_size || len % erase_size) {
		debug("SF: Erase offset/length not multiple of erase size\n");
		return -1;
	}

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: unable to claim SPI bus\n");
		return ret;
	}

	if (flash->flash_is_locked) {
		if (flash->flash_is_locked(flash, offset, len) > 0) {
			printf("offset 0x%x is protected and cannot be erased\n",
			       offset);
			ret = -EINVAL;
			goto release;
		}
	}

	while (len) {
		erase_addr = offset;

#ifdef CONFIG_SF_DUAL_FLASH
		if (flash->dual_flash > SF_SINGLE_FLASH)
			spi_flash_dual(flash, &erase_addr);
#endif
#ifdef CONFIG_SPI_FLASH_BAR
		ret = spi_flash_write_bar(flash, erase_addr);
		if (ret < 0)
			break;
#endif
		ret = spi_flash_cmd_write_enable(flash);
		if (ret < 0) {
			debug("SF: enabling write failed\n");
			break;
		}

		debug("SF: erase %2x %06x\n", flash->erase_cmd, erase_addr);
		ret = erase_fn(flash, erase_addr);
		if (ret < 0) {
			debug("SF: erase failed\n");
			break;
		}

		ret = spi_flash_cmd_wait_ready(flash,
					       SPI_FLASH_PAGE_ERASE_TIMEOUT);
		if (ret < 0) {
			debug("SF: write page erase timed out\n");
			break;
		}

		offset += erase_size;
		len -= erase_size;
	}

release:
	spi_release_bus(spi);

	return ret;
}

static int spi_flash_erase_impl(struct spi_flash *flash, u32 offset)
{
	struct spi_slave *spi = flash->spi;
	u8 cmd[SPI_FLASH_CMD_LEN];
	size_t cmdsz = 1 + flash->addr_width;

	cmd[0] = flash->erase_cmd;
	spi_flash_addr(flash, offset, cmd);
	return spi_flash_cmd_write(spi, cmd, cmdsz, NULL, 0);
}

int spi_flash_cmd_erase_ops(struct spi_flash *flash, u32 offset, size_t len)
{
	return spi_flash_erase_alg(flash, offset, len, spi_flash_erase_impl);
}

#if defined(CONFIG_SPI_FLASH_SST)
static int sst_write_wp(struct spi_flash *flash, u32 offset, size_t len,
			const void *buf, spi_flash_write_fn write_fn);

static int sst_write_bp(struct spi_flash *flash, u32 offset, size_t len,
			const void *buf, spi_flash_write_fn write_fn);
#endif

int spi_flash_write_alg(struct spi_flash *flash, u32 offset, size_t len,
			const void *buf, spi_flash_write_fn write_fn)
{
	struct spi_slave *spi = flash->spi;
	unsigned long byte_addr, page_size;
	u32 write_addr;
	size_t chunk_len, actual;
	int ret = -1;

#if defined(CONFIG_SPI_FLASH_SST)
	if (flash->flags & SNOR_F_SST_WR) {
		if (spi->mode & SPI_TX_BYTE)
			return sst_write_bp(flash, offset, len, buf, write_fn);
		else
			return sst_write_wp(flash, offset, len, buf, write_fn);
	}
#endif

	page_size = flash->page_size;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: unable to claim SPI bus\n");
		return ret;
	}

	if (flash->flash_is_locked) {
		if (flash->flash_is_locked(flash, offset, len) > 0) {
			printf("offset 0x%x is protected and cannot be written\n",
			       offset);
			ret = -EINVAL;
			goto release;
		}
	}

	for (actual = 0; actual < len; actual += chunk_len) {
		write_addr = offset;

#ifdef CONFIG_SF_DUAL_FLASH
		if (flash->dual_flash > SF_SINGLE_FLASH)
			spi_flash_dual(flash, &write_addr);
#endif
#ifdef CONFIG_SPI_FLASH_BAR
		ret = spi_flash_write_bar(flash, write_addr);
		if (ret < 0)
			break;
#endif
		byte_addr = offset % page_size;
		chunk_len = min(len - actual, (size_t)(page_size - byte_addr));

		if (spi->max_write_size)
			chunk_len = min(chunk_len,
					(size_t)spi->max_write_size);

		ret = spi_flash_cmd_write_enable(flash);
		if (ret < 0) {
			debug("SF: enabling write failed\n");
			break;
		}

		debug("SF: 0x%p => cmd = { 0x%02x 0x%06x } chunk_len = %zu\n",
		      buf + actual, flash->write_cmd, write_addr, chunk_len);

		ret = write_fn(flash, write_addr, chunk_len, buf + actual);
		if (ret < 0) {
			debug("SF: write failed\n");
			break;
		}

		ret = spi_flash_cmd_wait_ready(flash, SPI_FLASH_PROG_TIMEOUT);
		if (ret < 0) {
			debug("SF: write program timed out\n");
			break;
		}

		offset += chunk_len;
	}

release:
	spi_release_bus(spi);

	return ret;
}

static int spi_flash_write_impl(struct spi_flash *flash, u32 offset,
				size_t len, const void *buf)
{
	struct spi_slave *spi = flash->spi;
	u8 cmd[SPI_FLASH_CMD_LEN];
	size_t cmdsz = 1 + flash->addr_width;

	cmd[0] = flash->write_cmd;
	spi_flash_addr(flash, offset, cmd);
#ifdef CONFIG_SPI_FLASH_SST
	if (flash->flags & SNOR_F_SST_WR_2ND)
		cmdsz = 1;
#endif
	return spi_flash_cmd_write(spi, cmd, cmdsz, buf, len);
}

int spi_flash_cmd_write_ops(struct spi_flash *flash, u32 offset,
			    size_t len, const void *buf)
{
	return spi_flash_write_alg(flash, offset, len, buf,
				   spi_flash_write_impl);
}

int spi_flash_cmd_read_reg_ops(struct spi_flash *flash, u8 opcode,
			       size_t len, void *buf)
{
	struct spi_slave *spi = flash->spi;

	return spi_flash_cmd_read(spi, &opcode, 1, buf, len);
}

/*
 * TODO: remove the weak after all the other spi_flash_copy_mmap
 * implementations removed from drivers
 */
void __weak spi_flash_copy_mmap(void *data, void *offset, size_t len)
{
#ifdef CONFIG_DMA
	if (!dma_memcpy(data, offset, len))
		return;
#endif
	memcpy(data, offset, len);
}

int spi_flash_read_alg(struct spi_flash *flash, u32 offset, size_t len,
		       void *data, spi_flash_read_fn read_fn)
{
	struct spi_slave *spi = flash->spi;
	u32 remain_len, read_len, read_addr;
	int bank_sel = 0;
	int ret = -1;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: unable to claim SPI bus\n");
		return ret;
	}

	/* Handle memory-mapped SPI */
	if (flash->memory_map) {
		spi_xfer(spi, 0, NULL, NULL, SPI_XFER_MMAP);
		spi_flash_copy_mmap(data, flash->memory_map + offset, len);
		spi_xfer(spi, 0, NULL, NULL, SPI_XFER_MMAP_END);
		goto release;
	}

	while (len) {
		read_addr = offset;

#ifdef CONFIG_SF_DUAL_FLASH
		if (flash->dual_flash > SF_SINGLE_FLASH)
			spi_flash_dual(flash, &read_addr);
#endif
#ifdef CONFIG_SPI_FLASH_BAR
		ret = spi_flash_write_bar(flash, read_addr);
		if (ret < 0)
			break;
		bank_sel = flash->bank_curr;
#endif
		remain_len = ((SPI_FLASH_16MB_BOUN << flash->shift) *
				(bank_sel + 1)) - offset;
		if (len < remain_len)
			read_len = len;
		else
			read_len = remain_len;

		ret = read_fn(flash, read_addr, read_len, data);
		if (ret < 0) {
			debug("SF: read failed\n");
			break;
		}

		offset += read_len;
		len -= read_len;
		data += read_len;
	}

release:
	spi_release_bus(spi);

	return ret;
}

static int spi_flash_read_impl(struct spi_flash *flash, u32 offset,
			       size_t len, void *buf)
{
	struct spi_slave *spi = flash->spi;
	u8 *cmd;
	size_t cmdsz;
	int ret;

	cmdsz = 1 + flash->addr_width + flash->dummy_byte;
	cmd = calloc(1, cmdsz);
	if (!cmd) {
		debug("SF: Failed to allocate cmd\n");
		return -ENOMEM;
	}

	cmd[0] = flash->read_cmd;
	spi_flash_addr(flash, offset, cmd);
	ret = spi_flash_cmd_read(spi, cmd, cmdsz, buf, len);
	free(cmd);
	return ret;
}

int spi_flash_cmd_read_ops(struct spi_flash *flash, u32 offset,
		size_t len, void *buf)
{
	return spi_flash_read_alg(flash, offset, len, buf, spi_flash_read_impl);
}

#ifdef CONFIG_SPI_FLASH_SST
static int sst_byte_write(struct spi_flash *flash, u32 offset, const void *buf,
			  spi_flash_write_fn write_fn)
{
	struct spi_slave *spi = flash->spi;
	int ret;

	flash->write_cmd = CMD_SST_BP;
	debug("BP[%02x]: 0x%p => cmd = { 0x%02x 0x%06x }\n",
	      spi_w8r8(spi, CMD_READ_STATUS), buf, flash->write_cmd, offset);

	ret = spi_flash_cmd_write_enable(flash);
	if (ret)
		return ret;

	ret = write_fn(flash, offset, 1, buf);
	if (ret)
		return ret;

	return spi_flash_cmd_wait_ready(flash, SPI_FLASH_PROG_TIMEOUT);
}

static int sst_write_wp(struct spi_flash *flash, u32 offset, size_t len,
			const void *buf, spi_flash_write_fn write_fn)
{
	struct spi_slave *spi = flash->spi;
	size_t actual;
	int ret;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: Unable to claim SPI bus\n");
		return ret;
	}

	flash->flags &= ~SNOR_F_SST_WR_2ND;

	/* If the data is not word aligned, write out leading single byte */
	actual = offset % 2;
	if (actual) {
		ret = sst_byte_write(flash, offset, buf, write_fn);
		if (ret)
			goto done;
	}
	offset += actual;

	ret = spi_flash_cmd_write_enable(flash);
	if (ret)
		goto done;

	flash->write_cmd = CMD_SST_AAI_WP;

	for (; actual < len - 1; actual += 2) {
		debug("WP[%02x]: 0x%p => cmd = { 0x%02x 0x%06x }\n",
		      spi_w8r8(spi, CMD_READ_STATUS), buf + actual,
		      flash->write_cmd, offset);

		ret = write_fn(flash, offset, 2, buf + actual);
		if (ret) {
			debug("SF: sst word program failed\n");
			break;
		}

		ret = spi_flash_cmd_wait_ready(flash, SPI_FLASH_PROG_TIMEOUT);
		if (ret)
			break;

		flash->flags |= SNOR_F_SST_WR_2ND;
		offset += 2;
	}
	flash->flags &= ~SNOR_F_SST_WR_2ND;

	if (!ret)
		ret = spi_flash_cmd_write_disable(flash);

	/* If there is a single trailing byte, write it out */
	if (!ret && actual != len)
		ret = sst_byte_write(flash, offset, buf + actual, write_fn);

 done:
	debug("SF: sst: program %s %zu bytes @ 0x%zx\n",
	      ret ? "failure" : "success", len, offset - actual);

	spi_release_bus(spi);
	return ret;
}

static int sst_write_bp(struct spi_flash *flash, u32 offset, size_t len,
			const void *buf, spi_flash_write_fn write_fn)
{
	struct spi_slave *spi = flash->spi;
	size_t actual;
	int ret;

	ret = spi_claim_bus(spi);
	if (ret) {
		debug("SF: Unable to claim SPI bus\n");
		return ret;
	}

	for (actual = 0; actual < len; actual++) {
		ret = sst_byte_write(flash, offset, buf + actual, write_fn);
		if (ret) {
			debug("SF: sst byte program failed\n");
			break;
		}
		offset++;
	}

	if (!ret)
		ret = spi_flash_cmd_write_disable(flash);

	debug("SF: sst: program %s %zu bytes @ 0x%zx\n",
	      ret ? "failure" : "success", len, offset - actual);

	spi_release_bus(spi);
	return ret;
}
#endif

#if defined(CONFIG_SPI_FLASH_STMICRO) || defined(CONFIG_SPI_FLASH_SST)
static void stm_get_locked_range(struct spi_flash *flash, u8 sr, loff_t *ofs,
				 u64 *len)
{
	u8 mask = SR_BP2 | SR_BP1 | SR_BP0;
	int shift = ffs(mask) - 1;
	int pow;

	if (!(sr & mask)) {
		/* No protection */
		*ofs = 0;
		*len = 0;
	} else {
		pow = ((sr & mask) ^ mask) >> shift;
		*len = flash->size >> pow;
		*ofs = flash->size - *len;
	}
}

/*
 * Return 1 if the entire region is locked, 0 otherwise
 */
static int stm_is_locked_sr(struct spi_flash *flash, loff_t ofs, u64 len,
			    u8 sr)
{
	loff_t lock_offs;
	u64 lock_len;

	stm_get_locked_range(flash, sr, &lock_offs, &lock_len);

	return (ofs + len <= lock_offs + lock_len) && (ofs >= lock_offs);
}

/*
 * Check if a region of the flash is (completely) locked. See stm_lock() for
 * more info.
 *
 * Returns 1 if entire region is locked, 0 if any portion is unlocked, and
 * negative on errors.
 */
int stm_is_locked(struct spi_flash *flash, u32 ofs, size_t len)
{
	int status;
	u8 sr;

	status = read_sr(flash, &sr);
	if (status < 0)
		return status;

	return stm_is_locked_sr(flash, ofs, len, sr);
}

/*
 * Lock a region of the flash. Compatible with ST Micro and similar flash.
 * Supports only the block protection bits BP{0,1,2} in the status register
 * (SR). Does not support these features found in newer SR bitfields:
 *   - TB: top/bottom protect - only handle TB=0 (top protect)
 *   - SEC: sector/block protect - only handle SEC=0 (block protect)
 *   - CMP: complement protect - only support CMP=0 (range is not complemented)
 *
 * Sample table portion for 8MB flash (Winbond w25q64fw):
 *
 *   SEC  |  TB   |  BP2  |  BP1  |  BP0  |  Prot Length  | Protected Portion
 *  --------------------------------------------------------------------------
 *    X   |   X   |   0   |   0   |   0   |  NONE         | NONE
 *    0   |   0   |   0   |   0   |   1   |  128 KB       | Upper 1/64
 *    0   |   0   |   0   |   1   |   0   |  256 KB       | Upper 1/32
 *    0   |   0   |   0   |   1   |   1   |  512 KB       | Upper 1/16
 *    0   |   0   |   1   |   0   |   0   |  1 MB         | Upper 1/8
 *    0   |   0   |   1   |   0   |   1   |  2 MB         | Upper 1/4
 *    0   |   0   |   1   |   1   |   0   |  4 MB         | Upper 1/2
 *    X   |   X   |   1   |   1   |   1   |  8 MB         | ALL
 *
 * Returns negative on errors, 0 on success.
 */
int stm_lock(struct spi_flash *flash, u32 ofs, size_t len)
{
	u8 status_old, status_new;
	u8 mask = SR_BP2 | SR_BP1 | SR_BP0;
	u8 shift = ffs(mask) - 1, pow, val;
	int ret;

	ret = read_sr(flash, &status_old);
	if (ret < 0)
		return ret;

	/* SPI NOR always locks to the end */
	if (ofs + len != flash->size) {
		/* Does combined region extend to end? */
		if (!stm_is_locked_sr(flash, ofs + len, flash->size - ofs - len,
				      status_old))
			return -EINVAL;
		len = flash->size - ofs;
	}

	/*
	 * Need smallest pow such that:
	 *
	 *   1 / (2^pow) <= (len / size)
	 *
	 * so (assuming power-of-2 size) we do:
	 *
	 *   pow = ceil(log2(size / len)) = log2(size) - floor(log2(len))
	 */
	pow = ilog2(flash->size) - ilog2(len);
	val = mask - (pow << shift);
	if (val & ~mask)
		return -EINVAL;

	/* Don't "lock" with no region! */
	if (!(val & mask))
		return -EINVAL;

	status_new = (status_old & ~mask) | val;

	/* Only modify protection if it will not unlock other areas */
	if ((status_new & mask) <= (status_old & mask))
		return -EINVAL;

	write_sr(flash, status_new);

	return 0;
}

/*
 * Unlock a region of the flash. See stm_lock() for more info
 *
 * Returns negative on errors, 0 on success.
 */
int stm_unlock(struct spi_flash *flash, u32 ofs, size_t len)
{
	uint8_t status_old, status_new;
	u8 mask = SR_BP2 | SR_BP1 | SR_BP0;
	u8 shift = ffs(mask) - 1, pow, val;
	int ret;

	ret = read_sr(flash, &status_old);
	if (ret < 0)
		return ret;

	/* Cannot unlock; would unlock larger region than requested */
	if (stm_is_locked_sr(flash, ofs - flash->erase_size, flash->erase_size,
			     status_old))
		return -EINVAL;
	/*
	 * Need largest pow such that:
	 *
	 *   1 / (2^pow) >= (len / size)
	 *
	 * so (assuming power-of-2 size) we do:
	 *
	 *   pow = floor(log2(size / len)) = log2(size) - ceil(log2(len))
	 */
	pow = ilog2(flash->size) - order_base_2(flash->size - (ofs + len));
	if (ofs + len == flash->size) {
		val = 0; /* fully unlocked */
	} else {
		val = mask - (pow << shift);
		/* Some power-of-two sizes are not supported */
		if (val & ~mask)
			return -EINVAL;
	}

	status_new = (status_old & ~mask) | val;

	/* Only modify protection if it will not lock other areas */
	if ((status_new & mask) >= (status_old & mask))
		return -EINVAL;

	write_sr(flash, status_new);

	return 0;
}
#endif

int spi_flash_set_dummy_byte(struct spi_flash *flash, u8 num_dummy_cycles)
{
	u8 dummy_bits = num_dummy_cycles;

	dummy_bits *= SPI_FLASH_PROTO_ADR_FROM_PROTO(flash->read_proto);
	if (dummy_bits & 0x7)
		return -EINVAL;

	flash->dummy_byte = dummy_bits >> 3;
	return 0;
}

#if defined(CONFIG_SPI_FLASH_SPANSION) || defined(CONFIG_SPI_FLASH_WINBOND)
static int spansion_quad_enable(struct spi_flash *flash)
{
	u8 qeb_status;
	int ret;

	ret = read_cr(flash, &qeb_status);
	if (ret < 0)
		return ret;

	if (qeb_status & STATUS_QEB_WINSPAN)
		return 0;

	ret = write_cr(flash, qeb_status | STATUS_QEB_WINSPAN);
	if (ret < 0)
		return ret;

	/* read CR and check it */
	ret = read_cr(flash, &qeb_status);
	if (!(ret >= 0 && (qeb_status & STATUS_QEB_WINSPAN))) {
		printf("SF: Spansion CR Quad bit not clear\n");
		return -EINVAL;
	}

	return ret;
}
#endif

static int set_quad_mode(struct spi_flash *flash, u8 idcode0)
{
	switch (idcode0) {
#if defined(CONFIG_SPI_FLASH_SPANSION) || defined(CONFIG_SPI_FLASH_WINBOND)
	case SPI_FLASH_CFI_MFR_SPANSION:
	case SPI_FLASH_CFI_MFR_WINBOND:
		return spansion_quad_enable(flash);
#endif
	default:
		printf("SF: Need set QEB func for %02x flash\n", idcode0);
		return -1;
	}
}

/* Transitional function to be removed */
static int spi_flash_setup_deprecated(struct spi_flash *flash,
				      const struct spi_flash_params *params,
				      int best_match)
{
	struct spi_slave *spi = flash->spi;
	u8 idcode0 = params->jedec >> 16;
	u8 cmd;
	int ret;
	static u8 spi_read_cmds_array[] = {
		CMD_READ_ARRAY_SLOW,
		CMD_READ_ARRAY_FAST,
		CMD_READ_DUAL_OUTPUT_FAST,
		CMD_READ_QUAD_OUTPUT_FAST,
		CMD_READ_DUAL_IO_FAST,
		CMD_READ_QUAD_IO_FAST,
		CMD_READ_DUAL_IO_FAST,	/* same op code as for DUAL_IO_FAST */
		CMD_READ_QUAD_IO_FAST,	/* same op code as for QUAD_IO_FAST */
	};

	/* Look for the fastest read cmd */
	if (best_match) {
		cmd = spi_read_cmds_array[best_match - 1];
		flash->read_cmd = cmd;
	} else {
		/* Go for default supported read cmd */
		flash->read_cmd = CMD_READ_ARRAY_FAST;
	}

	/* Not require to look for fastest only two write cmds yet */
	if (params->flags & WR_QPP && spi->mode & SPI_TX_QUAD)
		flash->write_cmd = CMD_QUAD_PAGE_PROGRAM;
	else
		/* Go for default supported write cmd */
		flash->write_cmd = CMD_PAGE_PROGRAM;

	/* Set the quad enable bit - only for quad commands */
	if ((flash->read_cmd == CMD_READ_QUAD_OUTPUT_FAST) ||
	    (flash->read_cmd == CMD_READ_QUAD_IO_FAST) ||
	    (flash->write_cmd == CMD_QUAD_PAGE_PROGRAM)) {
		ret = set_quad_mode(flash, idcode0);
		if (ret) {
			debug("SF: Fail to set QEB for %02x\n", idcode0);
			return -EINVAL;
		}
	}

	/* Read dummy_byte: dummy byte is determined based on the
	 * dummy cycles of a particular command.
	 * Fast commands - dummy_byte = dummy_cycles/8
	 * I/O commands- dummy_byte = (dummy_cycles * no.of lines)/8
	 * For I/O commands except cmd[0] everything goes on no.of lines
	 * based on particular command but incase of fast commands except
	 * data all go on single line irrespective of command.
	 */
	switch (flash->read_cmd) {
	case CMD_READ_QUAD_IO_FAST:
		flash->dummy_byte = 2;
		break;
	case CMD_READ_ARRAY_SLOW:
		flash->dummy_byte = 0;
		break;
	default:
		flash->dummy_byte = 1;
	}

	return 0;
}

static int spi_flash_setup(struct spi_flash *flash,
			   const struct spi_flash_params *params,
			   u8 e_rd_cmd)
{
	int match = fls(params->e_rd_cmd & e_rd_cmd);

	/*
	 * The (Fast) Read and Page Program opcodes to be used depends on the
	 * manufacturer.
	 */
	switch (params->jedec >> 16) {
#ifdef CONFIG_SPI_FLASH_STMICRO
	case SPI_FLASH_CFI_MFR_STMICRO:
		return spi_flash_setup_micron(flash, params, match);
#endif

#ifdef CONFIG_SPI_FLASH_MACRONIX
	case SPI_FLASH_CFI_MFR_MACRONIX:
		return spi_flash_setup_macronix(flash, params, match);
#endif

	default:
		return spi_flash_setup_deprecated(flash, params, match);
	}

	return 0;
}

#if CONFIG_IS_ENABLED(OF_CONTROL)
int spi_flash_decode_fdt(const void *blob, struct spi_flash *flash)
{
#ifdef CONFIG_DM_SPI_FLASH
	fdt_addr_t addr;
	fdt_size_t size;
	int node = flash->dev->of_offset;

	addr = fdtdec_get_addr_size(blob, node, "memory-map", &size);
	if (addr == FDT_ADDR_T_NONE) {
		debug("%s: Cannot decode address\n", __func__);
		return 0;
	}

	if (flash->size != size) {
		debug("%s: Memory map must cover entire device\n", __func__);
		return -1;
	}
	flash->memory_map = map_sysmem(addr, size);
#endif

	return 0;
}
#endif /* CONFIG_IS_ENABLED(OF_CONTROL) */

struct spi_flash_read_id_config {
	enum spi_flash_protocol	proto;
	u8			e_rd_cmd;
	u8			cmd;
	int			idlen;
};

static const struct spi_flash_read_id_config configs[] = {
	/* Regular JEDEC Read ID (MUST be first, always tested) */
	{SPI_FLASH_PROTO_1_1_1, (ARRAY_SLOW | ARRAY_FAST), CMD_READ_ID, 5},
#if defined(CONFIG_SPI_FLASH_WINBOND)
	/* Winbond QPI mode */
	{SPI_FLASH_PROTO_4_4_4, QUAD_CMD_FAST, CMD_READ_ID, 5},
#endif
#if defined(CONFIG_SPI_FLASH_STMICRO) || defined(CONFIG_SPI_FLASH_MACRONIX)
	/* Micron Quad mode & Macronix QPI mode */
	{SPI_FLASH_PROTO_4_4_4, QUAD_CMD_FAST, CMD_READ_ID_MIO, 3},
#endif
#if defined(CONFIG_SPI_FLASH_STMICRO)
	/* Micron Dual mode */
	{SPI_FLASH_PROTO_2_2_2, DUAL_CMD_FAST, CMD_READ_ID_MIO, 3},
#endif
	/* Sentinel */
	{SPI_FLASH_PROTO_1_1_1, 0, 0, 0},
};

int spi_flash_scan(struct spi_flash *flash, u8 e_rd_cmd)
{
	struct spi_slave *spi = flash->spi;
	const struct spi_flash_params *params;
	const struct spi_flash_read_id_config *cfg;
	u16 jedec, ext_jedec;
	u8 idcode[5];
	int ret;

	/* Assign spi_flash ops */
#ifndef CONFIG_DM_SPI_FLASH
	flash->read_reg = spi_flash_cmd_read_reg_ops;
	flash->write_reg = spi_flash_cmd_write_reg_ops;
	flash->write = spi_flash_cmd_write_ops;
	flash->erase = spi_flash_cmd_erase_ops;
	flash->read = spi_flash_cmd_read_ops;
#endif

	/*
	 * Check whether the SPI NOR memory has already been configured (at
	 * reset of by some bootloader) to use a protocol other than SPI 1-1-1.
	 */
	params = NULL;
	jedec = 0;
	ext_jedec = 0;
	for (cfg = configs; cfg->e_rd_cmd; ++cfg) {
		/* Only try protocols supported by the SPI controller */
		if (cfg != configs && !(e_rd_cmd & cfg->e_rd_cmd))
			continue;

		/* Set this protocol for all commands */
		flash->reg_proto = cfg->proto;
		flash->read_proto = cfg->proto;
		flash->write_proto = cfg->proto;
		flash->erase_proto = cfg->proto;

		/* Read the ID codes */
		memset(idcode, 0, sizeof(idcode));
		ret = spi_flash_read_reg(flash, cfg->cmd, cfg->idlen, idcode);
		if (ret) {
			printf("SF: Failed to get idcodes\n");
			return -EINVAL;
		}

#ifdef DEBUG
		printf("SF: Got idcodes\n");
		print_buffer(0, idcode, 1, sizeof(idcode), 0);
#endif

		jedec = idcode[1] << 8 | idcode[2];
		ext_jedec = idcode[3] << 8 | idcode[4];

		/* Validate params from spi_flash_params table */
		params = spi_flash_params_table;
		for (; params->name != NULL; params++) {
			if ((params->jedec >> 16) == idcode[0]) {
				if ((params->jedec & 0xFFFF) == jedec) {
					if (params->ext_jedec == 0)
						break;
					else if (params->ext_jedec == ext_jedec)
						break;
				}
			}
		}

		if (params->name)
			break;
	}

	if (!params->name) {
		printf("SF: Unsupported flash IDs: ");
		printf("manuf %02x, jedec %04x, ext_jedec %04x\n",
		       idcode[0], jedec, ext_jedec);
		return -EPROTONOSUPPORT;
	}

	/* Flash powers up read-only, so clear BP# bits */
	if (idcode[0] == SPI_FLASH_CFI_MFR_ATMEL ||
	    idcode[0] == SPI_FLASH_CFI_MFR_MACRONIX ||
	    idcode[0] == SPI_FLASH_CFI_MFR_SST)
		write_sr(flash, 0);

	/* Assign spi data */
	flash->name = params->name;
	flash->memory_map = spi->memory_map;
	flash->dual_flash = spi->option;

	/* Assign spi flash flags */
	if (params->flags & SST_WR)
		flash->flags |= SNOR_F_SST_WR;

	/* lock hooks are flash specific - assign them based on idcode0 */
	switch (idcode[0]) {
#if defined(CONFIG_SPI_FLASH_STMICRO) || defined(CONFIG_SPI_FLASH_SST)
	case SPI_FLASH_CFI_MFR_STMICRO:
	case SPI_FLASH_CFI_MFR_SST:
		flash->flash_lock = stm_lock;
		flash->flash_unlock = stm_unlock;
		flash->flash_is_locked = stm_is_locked;
#endif
		break;
	default:
		debug("SF: Lock ops not supported for %02x flash\n", idcode[0]);
	}

	/* Compute the flash size */
	flash->shift = (flash->dual_flash & SF_DUAL_PARALLEL_FLASH) ? 1 : 0;
	/*
	 * The Spansion S25FL032P and S25FL064P have 256b pages, yet use the
	 * 0x4d00 Extended JEDEC code. The rest of the Spansion flashes with
	 * the 0x4d00 Extended JEDEC code have 512b pages. All of the others
	 * have 256b pages.
	 */
	if (ext_jedec == 0x4d00) {
		if ((jedec == 0x0215) || (jedec == 0x216))
			flash->page_size = 256;
		else
			flash->page_size = 512;
	} else {
		flash->page_size = 256;
	}
	flash->page_size <<= flash->shift;
	flash->sector_size = params->sector_size << flash->shift;
	flash->size = flash->sector_size * params->nr_sectors << flash->shift;
#ifdef CONFIG_SF_DUAL_FLASH
	if (flash->dual_flash & SF_DUAL_STACKED_FLASH)
		flash->size <<= 1;
#endif

	/* Compute erase sector and command */
	if (params->flags & SECT_4K) {
		flash->erase_cmd = CMD_ERASE_4K;
		flash->erase_size = 4096 << flash->shift;
	} else if (params->flags & SECT_32K) {
		flash->erase_cmd = CMD_ERASE_32K;
		flash->erase_size = 32768 << flash->shift;
	} else {
		flash->erase_cmd = CMD_ERASE_64K;
		flash->erase_size = flash->sector_size;
	}

	/* Now erase size becomes valid sector size */
	flash->sector_size = flash->erase_size;

	/*
	 * Setup (Fast) Read & Page Program opcode and protocols.
	 * Also set the right number of dummy/mode cycles.
	 */
	ret = spi_flash_setup(flash, params, e_rd_cmd);
	if (ret < 0)
		return ret;

#ifdef CONFIG_SPI_FLASH_STMICRO
	if (params->flags & E_FSR)
		flash->flags |= SNOR_F_USE_FSR;
#endif

	/* Configure the BAR - discover bank cmds and read current bank */
#ifdef CONFIG_SPI_FLASH_BAR
	ret = spi_flash_read_bar(flash, idcode[0]);
	if (ret < 0)
		return ret;
#endif

#if CONFIG_IS_ENABLED(OF_CONTROL)
	ret = spi_flash_decode_fdt(gd->fdt_blob, flash);
	if (ret) {
		debug("SF: FDT decode error\n");
		return -EINVAL;
	}
#endif

#ifndef CONFIG_SPL_BUILD
	printf("SF: Detected %s with page size ", flash->name);
	print_size(flash->page_size, ", erase size ");
	print_size(flash->erase_size, ", total ");
	print_size(flash->size, "");
	if (flash->memory_map)
		printf(", mapped at %p", flash->memory_map);
	puts("\n");
#endif

	/* Set number of bytes for address and convert opcodes if needed */
	flash->addr_width = 3;
	if (((flash->dual_flash == SF_SINGLE_FLASH) &&
	     (flash->size > SPI_FLASH_16MB_BOUN)) ||
	     ((flash->dual_flash > SF_SINGLE_FLASH) &&
	     (flash->size > SPI_FLASH_16MB_BOUN << 1))) {
#if defined(CONFIG_SPI_FLASH_4B_OPCODES)
		flash->addr_width = 4;
		flash->read_cmd  = spi_flash_3to4_opcode(flash->read_cmd);
		flash->write_cmd = spi_flash_3to4_opcode(flash->write_cmd);
		flash->erase_cmd = spi_flash_3to4_opcode(flash->erase_cmd);
#elif !defined(CONFIG_SPI_FLASH_BAR)
		puts("SF: Warning - Only lower 16MiB accessible,");
		puts(" Full access #define CONFIG_SPI_FLASH_ABOVE_16MB\n");
#endif
	}

	return ret;
}
